# -*- coding: utf-8 -*-
from typing import List, Dict, Any, NamedTuple, Union

import dataiku
from dataiku.customrecipe import get_recipe_config, get_input_names_for_role, get_output_names_for_role
from dkulib.dku_io_utils import set_column_descriptions
from dkulib.parallelizer import DataFrameParallelizer
from gpt_api_client import API_CATCH_EXCEPTIONS, GPTClient
from plugin_io_utils import ErrorHandlingEnum, validate_column_input, generate_unique
from text_classification import GPTZeroShotTextClassifier

# ==============================================================================
# SETUP
# ==============================================================================

recipe_config = get_recipe_config()

api_configuration_preset = recipe_config.get("api_configuration_preset")
if api_configuration_preset is None or api_configuration_preset == {}:
    raise ValueError("Please specify an API configuration preset")

# text_column, input_desc, classes and output_desc will trigger the following error if empty:
# <class 'ValueError'>: You must specify a valid column name
text_column = recipe_config.get("text_column")
input_desc = recipe_config.get("input_desc")
output_desc = recipe_config.get("output_desc", "")
explanation_enabled = recipe_config.get("explanation_enabled")

classes_str = recipe_config.get("classes")
# Note: putting class_name systematically before class_id seems to improve quality of predictions
classes: List[Dict[str, str]] = [{"class_name": class_name, "class_id": str(index+1)} for index, class_name in enumerate(classes_str)]

examples_map: Dict[str, str] = recipe_config.get("examples")
examples: List[Dict[str, Dict[str, Union[int, str]]]] = []
for index, example_text in enumerate(examples_map):
    example_class = examples_map[example_text]
    matching_classes = list(filter(lambda pred_class: pred_class["class_name"] == example_class, classes))
    if len(matching_classes) == 0:
        raise ValueError(
            f"Example number {index+1} has an invalid class \"{example_class}\""
        )
    else:
        examples.append({example_text: {"class_id": matching_classes[0]["class_id"]}})


column_prefix = "gpt"
parallel_workers = api_configuration_preset.get("parallel_workers")
error_handling = (
    ErrorHandlingEnum.FAIL if get_recipe_config().get("fail_on_error") else ErrorHandlingEnum.LOG
)

# Create client
client = GPTClient(
    api_configuration_preset=api_configuration_preset,
    temperature=recipe_config.get("temperature", 0),
    max_tokens=1024,  # Will be much shorter in practice since the output format is constrained but not too small (
    # when explanations are used)
)

input_dataset_name = get_input_names_for_role("input_dataset")[0]
# Not sure this check is needed
if not input_dataset_name:
    raise ValueError(
        "Cannot find input dataset. Please select an input dataset."
    )
input_dataset = dataiku.Dataset(input_dataset_name)
validate_column_input(text_column, [col["name"] for col in input_dataset.read_schema()])
input_df = input_dataset.get_dataframe()

output_column_name = generate_unique(
    name="prediction",
    existing_names=input_df.columns,
    prefix=column_prefix,
)
input_df[output_column_name] = ""
explanation_column_name = generate_unique(
    name="explanation",
    existing_names=input_df.columns,
    prefix=column_prefix,
)
if explanation_enabled:
    input_df[explanation_column_name] = ""

output_dataset_name = get_output_names_for_role("output_dataset")[0]
# Not sure this check is needed
if not output_dataset_name:
    raise ValueError(
        "Cannot find output dataset. Please select an output dataset."
    )
output_dataset = dataiku.Dataset(get_output_names_for_role("output_dataset")[0])

classifier = GPTZeroShotTextClassifier(
    gpt_client=client,
    input_desc=input_desc,
    classes=classes,
    output_desc=output_desc,
    examples=examples,
    explanation_enabled=explanation_enabled
)


def fill_values_from_response(batch: List[Dict], response: List[Any], output_column_names: NamedTuple) -> List[Dict]:
    return [
        {
            output_column_names.response: response.response_object,
            output_column_names.error_message: "",
            output_column_names.error_type: "",
            output_column_names.error_raw: "",
            **row,
            output_column_name: response.predicted_class,
            explanation_column_name: response.explanation
        }
        for response, row in zip(response, batch)
    ]


# ==============================================================================
# RUN
# ==============================================================================

df_parallelizer = DataFrameParallelizer(
    function=lambda row: classifier.get_response(input_text=str(row[text_column])),
    error_handling=error_handling,
    batch_response_parser=fill_values_from_response,
    exceptions_to_catch=API_CATCH_EXCEPTIONS,
    parallel_workers=parallel_workers,
    output_column_prefix=column_prefix,
)

output_df = df_parallelizer.run(input_df)
output_dataset.write_with_schema(output_df)

try:
    # Ugly. I feel better with this in a try catch
     gpt_desc_column_desc = {
         df_parallelizer._output_column_names._asdict()[name]: df_parallelizer.OUTPUT_COLUMN_NAME_DESCRIPTIONS[name]
         for name in df_parallelizer.OUTPUT_COLUMN_NAME_DESCRIPTIONS
     }
except:
    gpt_desc_column_desc = {}

set_column_descriptions(
    input_dataset=input_dataset,
    output_dataset=output_dataset,
    column_descriptions={
        output_column_name: output_desc,
        **gpt_desc_column_desc
    }
)
