# -*- coding: utf-8 -*-
from typing import Dict, NamedTuple
from typing import List

import dataiku
from dataiku.customrecipe import get_input_names_for_role
from dataiku.customrecipe import get_output_names_for_role
from dataiku.customrecipe import get_recipe_config
from dkulib.dku_io_utils import set_column_descriptions
from dkulib.parallelizer import DataFrameParallelizer
from gpt_api_client import API_CATCH_EXCEPTIONS, GPTClient
from plugin_io_utils import ErrorHandlingEnum, generate_unique
from plugin_io_utils import validate_column_input
from summarization import GPTTextSummarizer, GPTTextSummarizerResult

# ==============================================================================
# SETUP
# ==============================================================================

recipe_config = get_recipe_config()

api_configuration_preset = recipe_config.get("api_configuration_preset")
if api_configuration_preset is None or api_configuration_preset == {}:
    raise ValueError("Please specify an API configuration preset")

text_column = recipe_config.get("text_column")
summary_length_limit_enabled = recipe_config.get("summary_length_limit_enabled")
summary_length_unit = recipe_config.get("summary_length_unit")
summary_length_value = recipe_config.get("summary_length_value")


# Params for parallelization
column_prefix = "gpt"
parallel_workers = api_configuration_preset.get("parallel_workers")
error_handling = (
    ErrorHandlingEnum.FAIL if get_recipe_config().get("fail_on_error") else ErrorHandlingEnum.LOG
)

# Create client
client = GPTClient(
    api_configuration_preset=api_configuration_preset,
    temperature=recipe_config.get("temperature", 0),
    max_tokens=recipe_config.get("max_tokens", 64)
)

# ==============================================================================
# DEFINITIONS
# ==============================================================================


input_dataset_names = get_input_names_for_role("input_dataset")
if not input_dataset_names:
    raise ValueError(
        "Cannot find input dataset. Use Output-only mode to generate without input dataset."
    )
input_dataset = dataiku.Dataset(input_dataset_names[0])
validate_column_input(text_column, [col["name"] for col in input_dataset.read_schema()])
input_df = input_dataset.get_dataframe()
output_column_name = generate_unique(
    name="summary",
    existing_names=input_df.columns,
    prefix=column_prefix,
)
input_df[output_column_name] = ""

output_dataset = dataiku.Dataset(get_output_names_for_role("output_dataset")[0])
text_summarizer = GPTTextSummarizer(client, summary_length_limit_enabled, summary_length_unit, summary_length_value)


def fill_values_from_response(batch: List[Dict], response: List[GPTTextSummarizerResult],
                              output_column_names: NamedTuple) -> List[Dict]:
    return [
        {
            output_column_names.response: response.response_object,
            output_column_names.error_message: "",
            output_column_names.error_type: "",
            output_column_names.error_raw: "",
            **row,
            output_column_name: response.summarized_text
        }
        for response, row in zip(response, batch)
    ]


# ==============================================================================
# RUN
# ==============================================================================

df_parallelizer = DataFrameParallelizer(
    function=lambda row: text_summarizer.get_response(row[text_column]),
    error_handling=error_handling,
    batch_response_parser=fill_values_from_response,
    exceptions_to_catch=API_CATCH_EXCEPTIONS,
    parallel_workers=parallel_workers,
    output_column_prefix=column_prefix,
)

output_df = df_parallelizer.run(input_df)
output_dataset.write_with_schema(output_df)

try:
    # Ugly. I feel better with this in a try catch
     gpt_desc_column_desc = {
         df_parallelizer._output_column_names._asdict()[name]: df_parallelizer.OUTPUT_COLUMN_NAME_DESCRIPTIONS[name]
         for name in df_parallelizer.OUTPUT_COLUMN_NAME_DESCRIPTIONS
     }
except:
    gpt_desc_column_desc = {}

set_column_descriptions(
    input_dataset=input_dataset,
    output_dataset=output_dataset,
    column_descriptions={
        output_column_name: f'Summary of "{text_column}" column',
        **gpt_desc_column_desc
    }
)
