import logging
from typing import List, Tuple, Any

from gpt_api_client import GPTClient


class GPTTextGeneratorResult:
    def __init__(self, response_object, generated_text):
        self.response_object = response_object
        self.generated_text = generated_text


class GPTTextGenerator:
    def __init__(self, gpt_client: GPTClient, task: str, examples: List[Tuple[str, str]], input_desc: str,
                 output_desc: str, uses_input_text: bool):
        self.gpt_client = gpt_client
        self.task = task
        self.examples = examples
        self.input_desc = input_desc
        self.output_desc = output_desc
        self.uses_input_text = uses_input_text

        logging.info("GPT API client params:")
        logging.info(gpt_client)

        logging.info("Prompt used for text generation:")
        logging.info(self._format_prompt("[input text]"))

    def _format_prompt(self, text: str = "") -> str:
        """
        Returns prompt of form:

        Correct grammar mistakes.

        Original: Where do you went?
        Standard American English: Where did you go?
        Original: Where is you?
        Standard American English: Where are you?

        Args:
            task: The task for GPT, e.g. Correct grammar mistakes.
            text: The current row based on which to generate
            input_desc: Description of input column
            output_desc: Description of output column
            example_in: Example of input text
            example_out: Example of desired output text
        Returns:
            prompt: Formatted prompt
        """

        prompt = ""

        ### Task ###
        if self.task:
            prompt += f"{self.task}\n\n"

        ### Examples ###
        for ex_inp, ex_out in self.examples:
            if ex_inp:
                if self.input_desc:
                    prompt += f"{self.input_desc}: "
                prompt += f"{ex_inp}\n"

            # One could also provide output examples without descriptions, e.g.
            # elephant
            # giraffe
            # cat
            if ex_out:
                if self.output_desc:
                    prompt += f"{self.output_desc}: "
                prompt += f"{ex_out}\n"

        ### Final prompt ###
        if self.uses_input_text:
            if self.input_desc:
                prompt += f"{self.input_desc}: "
            prompt += f"{text}\n"

        if self.output_desc:
            # Do not end with a space, as it worsens generation
            prompt += f"{self.output_desc}:"

        return prompt

    def get_response(self, text: Any) -> GPTTextGeneratorResult:
        """
        Constructs a prompt and makes an API call to generate text.
        """
        prompt = self._format_prompt(str(text))
        gpt_response = self.gpt_client.send_prompt(prompt)

        return GPTTextGeneratorResult(
            gpt_response.raw_response,
            gpt_response.response_text
        )
